from asdf.extension import Converter

# These attributes don't end up in the hdulist and
# instead will be stored in "attrs"
_WCS_ATTRS = ("naxis", "colsel", "keysel", "key", "pixel_bounds")


class WCSConverter(Converter):
    tags = ("tag:astropy.org:astropy/wcs/wcs-*",)
    types = ("astropy.wcs.wcs.WCS",)

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.wcs import WCS

        hdulist = node["hdulist"]
        attrs = node["attrs"]

        if naxis := attrs.pop("naxis"):
            hdulist[0].header["naxis"] = naxis

        pixel_bounds = attrs.pop("pixel_bounds")

        wcs = WCS(hdulist[0].header, fobj=hdulist, **attrs)

        if wcs.sip is not None:
            # work around a bug in astropy where sip headers lose precision
            # see https://github.com/astropy/astropy/issues/17334
            wcs.sip = wcs._read_sip_kw(hdulist[0].header, attrs.get("key", " "))
            wcs.wcs.set()

        wcs.pixel_bounds = pixel_bounds
        return wcs

    def to_yaml_tree(self, wcs, tag, ctx):
        hdulist = wcs.to_fits(relax=True)
        attrs = {a: getattr(wcs, a) for a in _WCS_ATTRS if hasattr(wcs, a)}
        return {"hdulist": hdulist, "attrs": attrs}



import io
import sys
import warnings
from pathlib import Path

import asdf
import pytest
import yaml


def test_resources():
    resources_root = Path(__file__).parent.parent / "resources"
    resource_manager = asdf.get_config().resource_manager

    for resource_path in resources_root.glob("**/*.yaml"):
        with resource_path.open("rb") as f:
            resource_content = f.read()
        resource = yaml.safe_load(resource_content)
        resource_uri = resource["id"]
        assert resource_manager[resource_uri] == resource_content


def test_manifests():
    manifests_root = Path(__file__).parent.parent / "resources" / "manifests"
    resource_manager = asdf.get_config().resource_manager

    for manifest_path in manifests_root.glob("*.yaml"):
        with manifest_path.open("rb") as f:
            manifest_content = f.read()
        manifest = yaml.safe_load(manifest_content)

        manifest_schema = asdf.schema.load_schema("asdf://asdf-format.org/core/schemas/extension_manifest-1.0.0")

        # The manifest must be valid against its own schema:
        asdf.schema.validate(manifest, schema=manifest_schema)

        for tag_definition in manifest["tags"]:
            # The tag's schema must be available:
            assert tag_definition["schema_uri"] in resource_manager


def test_extensions():
    package_and_uri_pairs = {(e.package_name, e.extension_uri) for e in asdf.get_config().extensions}

    assert ("asdf-astropy", "asdf://astropy.org/astropy/extensions/astropy-1.0.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.0.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.1.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.2.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.3.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.4.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.5.0") in package_and_uri_pairs


_ASTROPY_MODULES = [
    "astropy.coordinates",
    "astropy.io",
    "astropy.modeling",
    "astropy.table",
    "astropy.time",
    "astropy.units",
]


@pytest.fixture
def _clean_astropy_imports():
    """Temporally unload all astropy modules used by asdf-astropy"""

    # If any astropy or asdf_astropy modules are already imported
    # remove them from sys.modules so we can later in this test check
    # if those modules were imported. We have to store these as
    # many submodules will reference or use types defined in other
    # submodules
    previous_modules = {}
    for name in sys.modules.copy():
        if name.startswith("asdf_astropy") or any(name.startswith(m) for m in _ASTROPY_MODULES):
            previous_modules[name] = sys.modules[name]
            del sys.modules[name]

    # Register a module finder that just raises an exception if
    # one of the tracked astropy modules is imported
    class _Finder:
        def find_spec(self, modulename, path=None, target=None):
            if any(modulename.startswith(m) for m in _ASTROPY_MODULES):
                msg = f"attempt to import astropy submodule({modulename}) during integration"
                raise Exception(msg)  # noqa: TRY002

    sys.meta_path.insert(0, _Finder())

    # Setup complete return the test
    yield

    # Restore the previously imported modules. This is necessary as other code
    # may have already used the modules we removed from the cache (astropy.constants
    # defines some quantities).
    for name, previous_module in previous_modules.items():
        sys.modules[name] = previous_module
    if isinstance(sys.meta_path[0], _Finder):
        sys.meta_path.pop(0)


@pytest.mark.usefixtures("_clean_astropy_imports")
def test_no_astropy_import():
    """
    Confirm that none of the ASDF plugins import astropy modules
    at import time.
    """

    from asdf_astropy import integration

    integration.get_resource_mappings()
    integration.get_extensions()

    assert not any(k for k in sys.modules if any(k.startswith(m) for m in _ASTROPY_MODULES))


def test_no_core_extension_overwrite():
    """
    Check that this package (even though it implements converters for core types)
    does not overwrite the core extension provided by asdf
    """
    import astropy.units as u

    # define a tree with a unit that will be serialized by this package
    # the tree itself will be serialized by asdf (which should result in
    # including asdf as a used extension)
    tree = {"meter": u.m}

    bio = io.BytesIO()
    asdf.AsdfFile(tree).write_to(bio)
    bio.seek(0)

    with asdf.open(bio) as af:
        packages = [ext["software"]["name"] for ext in af.tree["history"]["extensions"]]
        # check that both asdf and asdf-astropy are recorded as being used to
        # generate the file
        assert "asdf" in packages
        assert "asdf-astropy" in packages


def test_no_warnings_for_astropy_manifest_files():
    """
    asdf-astropy 0.5.0 wrote files with astropy.org manifests
    that were compounds of core and astropy types. These types are
    now handled by different extensions. This test checks that files
    generated with the no longer used astropy.org manifests open without
    warnings.
    """
    file_contents = io.BytesIO(
        b"""#ASDF 1.0.0
#ASDF_STANDARD 1.5.0
%YAML 1.1
%TAG ! tag:stsci.edu:asdf/
--- !core/asdf-1.1.0
asdf_library: !core/software-1.0.0 {author: The ASDF Developers, homepage: 'http://github.com/asdf-format/asdf',
  name: asdf, version: 3.3.0}
history:
  extensions:
  - !core/extension_metadata-1.0.0
    extension_class: asdf.extension._manifest.ManifestExtension
    extension_uri: asdf://asdf-format.org/core/extensions/core-1.5.0
    manifest_software: !core/software-1.0.0 {name: asdf_standard, version: 1.1.1}
    software: !core/software-1.0.0 {name: asdf, version: 3.3.0}
  - !core/extension_metadata-1.0.0
    extension_class: asdf_astropy._manifest.CompoundManifestExtension
    extension_uri: asdf://astropy.org/core/extensions/core-1.5.0
    software: !core/software-1.0.0 {name: asdf-astropy, version: 0.5.0}
u: !unit/unit-1.0.0 m
...""",
    )
    with warnings.catch_warnings():
        # make sure warnings are errors
        warnings.simplefilter("error")
        with asdf.open(file_contents) as af:
            assert af["u"]  # we don't care about the contents here



class UncertaintyConverter:
    tags = ("tag:astropy.org:astropy/nddata/uncertainty-*",)
    types = (
        "astropy.nddata.nduncertainty.StdDevUncertainty",
        "astropy.nddata.nduncertainty.UnknownUncertainty",
        "astropy.nddata.nduncertainty.VarianceUncertainty",
    )

    # Mapping of uncertainty class name (attribute of astropy.nddata)
    # and code "name" stored in the data file.
    # This will need to be separately versioned if the schema is updated.
    _class_name_to_code = {
        "StdDevUncertainty": "stddev",
        "UnknownUncertainty": "unknown",
        "VarianceUncertainty": "variance",
    }
    _class_code_to_name = {v: k for k, v in _class_name_to_code.items()}

    def from_yaml_tree(self, node, tag, ctx):
        import astropy.nddata

        class_name = self._class_code_to_name[node["name"]]

        return getattr(astropy.nddata, class_name)(node["array"], unit=node.get("unit"))

    def to_yaml_tree(self, obj, tag, ctx):
        node = {
            "name": self._class_name_to_code[obj.__class__.__name__],
            "array": obj.array,
        }

        if obj.unit is not None:
            node["unit"] = obj.unit

        return node



import asdf
import numpy as np
import pytest
from astropy import units as u
from astropy.utils.masked import Masked

from asdf_astropy.tests.versions import ASTROPY_GE_7_1

if ASTROPY_GE_7_1:
    from astropy.utils.masked import get_data_and_mask
else:
    pytest.skip(reason="MaskedQuantity support was added in astropy 7.1", allow_module_level=True)

MaskedQuantity = Masked(u.Quantity)


def test_masked_quantity(tmp_path):
    data = [1, 2, 3]
    mask = [False, False, True]
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["quantity"] = Masked(data, mask) * u.yottamole
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert isinstance(af["quantity"], MaskedQuantity)
        assert af["quantity"].unit == u.yottamole

        result_data, result_mask = get_data_and_mask(af["quantity"].value)
        np.testing.assert_array_equal(result_data, data)
        np.testing.assert_array_equal(result_mask, mask)



import warnings

import numpy as np
import pytest
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename, get_pkg_data_filenames
from astropy.wcs import WCS, DistortionLookupTable, FITSFixedWarning, Sip

from asdf_astropy.testing.helpers import assert_wcs_roundtrip

_astropy_test_header_filenames = list(get_pkg_data_filenames("tests/data/maps", "astropy.wcs", "*.hdr")) + list(
    get_pkg_data_filenames("tests/data/spectra", "astropy.wcs", "*.hdr"),
)

_astropy_test_fits_filenames = [
    get_pkg_data_filename(f"tests/data/{fn}", "astropy.wcs")
    for fn in [
        "ie6d07ujq_wcs.fits",
        "j94f05bgq_flt.fits",
        "sip.fits",
        "sip2.fits",
    ]
]


def create_empty_wcs():
    return WCS()


def create_wcs_with_attrs():
    wcs = WCS(naxis=3)
    wcs.pixel_shape = [100, 200, 300]
    wcs.pixel_bounds = [[11, 22], [33, 45], [55, 67]]
    return wcs


def create_sip_distortion_wcs():
    rng = np.random.default_rng(42)
    wcs = WCS(naxis=2)
    wcs.wcs.crval = [251.29, 57.58]
    wcs.wcs.cdelt = [1, 1]
    wcs.wcs.crpix = [507, 507]
    wcs.wcs.pc = np.array([[7.7e-6, 3.3e-5], [3.7e-5, -6.8e-6]])
    wcs._naxis = [1014, 1014]
    wcs.wcs.ctype = ["RA---TAN-SIP", "DEC--TAN-SIP"]

    # Generate random SIP coefficients
    a = rng.uniform(low=-1e-5, high=1e-5, size=(5, 5))
    b = rng.uniform(low=-1e-5, high=1e-5, size=(5, 5))

    # Assign SIP coefficients
    wcs.sip = Sip(a, b, None, None, wcs.wcs.crpix)
    wcs.wcs.set()

    return wcs


def create_tabular_wcs():
    # Creates a WCS object with distortion lookup tables
    img_world_wcs = WCS(naxis=2)
    img_world_wcs.wcs.crpix = 1, 1
    img_world_wcs.wcs.crval = 0, 0
    img_world_wcs.wcs.cdelt = 1, 1

    # Create maps with zero distortion except at one particular pixel
    x_dist_array = np.zeros((25, 25))
    x_dist_array[10, 20] = 0.5
    map_x = DistortionLookupTable(
        x_dist_array.astype(np.float32),
        (5, 10),
        (10, 20),
        (2, 2),
    )
    y_dist_array = np.zeros((25, 25))
    y_dist_array[10, 5] = 0.7
    map_y = DistortionLookupTable(
        y_dist_array.astype(np.float32),
        (5, 10),
        (10, 20),
        (3, 3),
    )

    img_world_wcs.cpdis1 = map_x
    img_world_wcs.cpdis2 = map_y

    return img_world_wcs


@pytest.mark.parametrize("version", ["1.5.0", "1.6.0"])
@pytest.mark.parametrize(
    "wcs_gen",
    [create_empty_wcs, create_wcs_with_attrs, create_tabular_wcs, create_sip_distortion_wcs],
)
def test_roundtrip(wcs_gen, tmp_path, version):
    wcs = wcs_gen()
    assert_wcs_roundtrip(wcs, tmp_path, version)


@pytest.mark.parametrize("fn", _astropy_test_header_filenames)
@pytest.mark.parametrize("version", ["1.5.0", "1.6.0"])
def test_astropy_data_header_roundtrip(fn, tmp_path, version):
    with open(fn) as f:
        header = f.read()

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=FITSFixedWarning)
        wcs = WCS(header)

    assert_wcs_roundtrip(wcs, tmp_path, version)


@pytest.mark.parametrize("fn", _astropy_test_fits_filenames)
@pytest.mark.parametrize("version", ["1.5.0", "1.6.0"])
def test_astropy_data_fits_roundtrip(fn, tmp_path, version):
    with fits.open(fn) as ff:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=FITSFixedWarning)
            wcs = WCS(ff[0].header, ff)

        assert_wcs_roundtrip(wcs, tmp_path, version)



from asdf.extension import Converter


class EarthLocationConverter(Converter):
    tags = ("tag:astropy.org:astropy/coordinates/earthlocation-*",)
    types = (
        "astropy.coordinates.earth.EarthLocation",
        "astropy.utils.masked.core.MaskedEarthLocation",
    )

    def to_yaml_tree(self, obj, tag, ctx):
        result = obj.info._represent_as_dict()
        result.pop("__class__", None)
        return result

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.coordinates.earth import EarthLocation

        return EarthLocation.info._construct_from_dict(node)



"""
This module builds all of the ASDF extensions which will be registered by `asdf_astropy.integration`,
via an ``entry-point`` in the ``pyproject.toml`` file.
"""

import importlib.metadata
from importlib.util import find_spec

import asdf
from asdf.extension import Extension, ManifestExtension

from .converters.coordinates.angle import AngleConverter, LatitudeConverter, LongitudeConverter
from .converters.coordinates.earth_location import EarthLocationConverter
from .converters.coordinates.frame import FrameConverter, LegacyICRSConverter
from .converters.coordinates.representation import RepresentationConverter
from .converters.coordinates.sky_coord import SkyCoordConverter
from .converters.coordinates.spectral_coord import SpectralCoordConverter
from .converters.fits.fits import AsdfFitsConverter, AstropyFitsConverter
from .converters.nddata.uncertainty import UncertaintyConverter
from .converters.table.table import AsdfTableConverter, AstropyTableConverter, ColumnConverter, NdarrayMixinConverter
from .converters.time.time import TimeConverter
from .converters.time.time_delta import TimeDeltaConverter
from .converters.transform.compound import CompoundConverter
from .converters.transform.core import SimpleTransformConverter
from .converters.transform.functional_models import ConstantConverter
from .converters.transform.mappings import IdentityConverter, RemapAxesConverter, UnitsMappingConverter
from .converters.transform.math_functions import MathFunctionsConverter
from .converters.transform.polynomial import OrthoPolynomialConverter, PolynomialConverter
from .converters.transform.projections import ProjectionConverter
from .converters.transform.properties import CompoundBoundingBoxConverter, ModelBoundingBoxConverter
from .converters.transform.rotations import Rotate3DConverter, RotationSequenceConverter
from .converters.transform.spline import SplineConverter
from .converters.transform.tabular import TabularConverter
from .converters.unit.equivalency import EquivalencyConverter
from .converters.unit.magunit import MagUnitConverter
from .converters.unit.quantity import QuantityConverter
from .converters.unit.unit import UnitConverter
from .converters.wcs.slicedwcs import SlicedWCSConverter
from .converters.wcs.wcs import WCSConverter

__all__ = [
    "ASTROPY_CONVERTERS",
    "ASTROPY_EXTENSIONS",
    "COORDINATES_CONVERTERS",
    "COORDINATES_EXTENSIONS",
    "CORE_CONVERTERS",
    "CORE_EXTENSIONS",
    "CORE_MANIFEST_URIS",
    "TRANSFORM_CONVERTERS",
    "TRANSFORM_EXTENSIONS",
    "TRANSFORM_MANIFEST_URIS",
    "UNIT_EXTENSIONS",
]

TRANSFORM_CONVERTERS = [
    # astropy.modeling.core
    CompoundConverter(),
    # astropy.modeling.functional_models
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/airy_disk2d-*"],
        "astropy.modeling.functional_models.AiryDisk2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/box1d-*"],
        "astropy.modeling.functional_models.Box1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/box2d-*"],
        "astropy.modeling.functional_models.Box2D",
    ),
    ConstantConverter(),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/disk2d-*"],
        "astropy.modeling.functional_models.Disk2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ellipse2d-*"],
        "astropy.modeling.functional_models.Ellipse2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/exponential1d-*"],
        "astropy.modeling.functional_models.Exponential1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/gaussian1d-*"],
        "astropy.modeling.functional_models.Gaussian1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/gaussian2d-*"],
        "astropy.modeling.functional_models.Gaussian2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/king_projected_analytic1d-*"],
        "astropy.modeling.functional_models.KingProjectedAnalytic1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/linear1d-*"],
        "astropy.modeling.functional_models.Linear1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/logarithmic1d-*"],
        "astropy.modeling.functional_models.Logarithmic1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/lorentz1d-*"],
        "astropy.modeling.functional_models.Lorentz1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/moffat1d-*"],
        "astropy.modeling.functional_models.Moffat1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/moffat2d-*"],
        "astropy.modeling.functional_models.Moffat2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/multiplyscale-*"],
        "astropy.modeling.functional_models.Multiply",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/planar2d-*"],
        "astropy.modeling.functional_models.Planar2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/redshift_scale_factor-*"],
        "astropy.modeling.functional_models.RedshiftScaleFactor",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ricker_wavelet1d-*"],
        "astropy.modeling.functional_models.RickerWavelet1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ricker_wavelet2d-*"],
        "astropy.modeling.functional_models.RickerWavelet2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/ring2d-*"],
        "astropy.modeling.functional_models.Ring2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/scale-*"],
        "astropy.modeling.functional_models.Scale",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/sersic1d-*"],
        "astropy.modeling.functional_models.Sersic1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/sersic2d-*"],
        "astropy.modeling.functional_models.Sersic2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/shift-*"],
        "astropy.modeling.functional_models.Shift",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/sine1d-*"],
        "astropy.modeling.functional_models.Sine1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/cosine1d-*"],
        "astropy.modeling.functional_models.Cosine1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/tangent1d-*"],
        "astropy.modeling.functional_models.Tangent1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/arcsine1d-*"],
        "astropy.modeling.functional_models.ArcSine1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/arccosine1d-*"],
        "astropy.modeling.functional_models.ArcCosine1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/arctangent1d-*"],
        "astropy.modeling.functional_models.ArcTangent1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/trapezoid1d-*"],
        "astropy.modeling.functional_models.Trapezoid1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/trapezoid_disk2d-*"],
        "astropy.modeling.functional_models.TrapezoidDisk2D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/voigt1d-*"],
        "astropy.modeling.functional_models.Voigt1D",
    ),
    # astropy.modeling.mappings
    IdentityConverter(),
    RemapAxesConverter(),
    # UnitsMapping is not represented here because
    # it is an astropy-specific transform and not
    # included in the ASDF transform extension.
    # astropy.modeling.math_functions
    MathFunctionsConverter(),
    # astropy.modeling.physical_models
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/blackbody-*"],
        "astropy.modeling.physical_models.BlackBody",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/drude1d-*"],
        "astropy.modeling.physical_models.Drude1D",
    ),
    # TODO: Implement NFW
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/plummer1d-*"],
        "astropy.modeling.physical_models.Plummer1D",
    ),
    # astropy.modeling.polynomial
    PolynomialConverter(),
    OrthoPolynomialConverter(),
    # SIP and InverseSIP are deliberately excluded because they
    # are FITS-specific and can be easily represented by a
    # simple combination of existing models.
    SplineConverter(),
    # astropy.modeling.powerlaws
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/broken_power_law1d-*"],
        "astropy.modeling.powerlaws.BrokenPowerLaw1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/exponential_cutoff_power_law1d-*"],
        "astropy.modeling.powerlaws.ExponentialCutoffPowerLaw1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/log_parabola1d-*"],
        "astropy.modeling.powerlaws.LogParabola1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/power_law1d-*"],
        "astropy.modeling.powerlaws.PowerLaw1D",
    ),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/smoothly_broken_power_law1d-*"],
        "astropy.modeling.powerlaws.SmoothlyBrokenPowerLaw1D",
    ),
    # astropy.modeling.projections
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/affine-*"],
        "astropy.modeling.projections.AffineTransformation2D",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/airy-*"],
        "astropy.modeling.projections.Pix2Sky_Airy",
        "astropy.modeling.projections.Sky2Pix_Airy",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/bonne_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_BonneEqualArea",
        "astropy.modeling.projections.Sky2Pix_BonneEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/cobe_quad_spherical_cube-*"],
        "astropy.modeling.projections.Pix2Sky_COBEQuadSphericalCube",
        "astropy.modeling.projections.Sky2Pix_COBEQuadSphericalCube",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_ConicEqualArea",
        "astropy.modeling.projections.Sky2Pix_ConicEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_equidistant-*"],
        "astropy.modeling.projections.Pix2Sky_ConicEquidistant",
        "astropy.modeling.projections.Sky2Pix_ConicEquidistant",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_orthomorphic-*"],
        "astropy.modeling.projections.Pix2Sky_ConicOrthomorphic",
        "astropy.modeling.projections.Sky2Pix_ConicOrthomorphic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/conic_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_ConicPerspective",
        "astropy.modeling.projections.Sky2Pix_ConicPerspective",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/cylindrical_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_CylindricalEqualArea",
        "astropy.modeling.projections.Sky2Pix_CylindricalEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/cylindrical_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_CylindricalPerspective",
        "astropy.modeling.projections.Sky2Pix_CylindricalPerspective",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/gnomonic-*"],
        "astropy.modeling.projections.Pix2Sky_Gnomonic",
        "astropy.modeling.projections.Sky2Pix_Gnomonic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/hammer_aitoff-*"],
        "astropy.modeling.projections.Pix2Sky_HammerAitoff",
        "astropy.modeling.projections.Sky2Pix_HammerAitoff",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/healpix-*"],
        "astropy.modeling.projections.Pix2Sky_HEALPix",
        "astropy.modeling.projections.Sky2Pix_HEALPix",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/healpix_polar-*"],
        "astropy.modeling.projections.Pix2Sky_HEALPixPolar",
        "astropy.modeling.projections.Sky2Pix_HEALPixPolar",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/mercator-*"],
        "astropy.modeling.projections.Pix2Sky_Mercator",
        "astropy.modeling.projections.Sky2Pix_Mercator",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/molleweide-*"],
        "astropy.modeling.projections.Pix2Sky_Molleweide",
        "astropy.modeling.projections.Sky2Pix_Molleweide",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/parabolic-*"],
        "astropy.modeling.projections.Pix2Sky_Parabolic",
        "astropy.modeling.projections.Sky2Pix_Parabolic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/plate_carree-*"],
        "astropy.modeling.projections.Pix2Sky_PlateCarree",
        "astropy.modeling.projections.Sky2Pix_PlateCarree",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/polyconic-*"],
        "astropy.modeling.projections.Pix2Sky_Polyconic",
        "astropy.modeling.projections.Sky2Pix_Polyconic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/quad_spherical_cube-*"],
        "astropy.modeling.projections.Pix2Sky_QuadSphericalCube",
        "astropy.modeling.projections.Sky2Pix_QuadSphericalCube",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/sanson_flamsteed-*"],
        "astropy.modeling.projections.Pix2Sky_SansonFlamsteed",
        "astropy.modeling.projections.Sky2Pix_SansonFlamsteed",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/slant_orthographic-*"],
        "astropy.modeling.projections.Pix2Sky_SlantOrthographic",
        "astropy.modeling.projections.Sky2Pix_SlantOrthographic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/slant_zenithal_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_SlantZenithalPerspective",
        "astropy.modeling.projections.Sky2Pix_SlantZenithalPerspective",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/stereographic-*"],
        "astropy.modeling.projections.Pix2Sky_Stereographic",
        "astropy.modeling.projections.Sky2Pix_Stereographic",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/tangential_spherical_cube-*"],
        "astropy.modeling.projections.Pix2Sky_TangentialSphericalCube",
        "astropy.modeling.projections.Sky2Pix_TangentialSphericalCube",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/zenithal_equal_area-*"],
        "astropy.modeling.projections.Pix2Sky_ZenithalEqualArea",
        "astropy.modeling.projections.Sky2Pix_ZenithalEqualArea",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/zenithal_equidistant-*"],
        "astropy.modeling.projections.Pix2Sky_ZenithalEquidistant",
        "astropy.modeling.projections.Sky2Pix_ZenithalEquidistant",
    ),
    ProjectionConverter(
        ["tag:stsci.edu:asdf/transform/zenithal_perspective-*"],
        "astropy.modeling.projections.Pix2Sky_ZenithalPerspective",
        "astropy.modeling.projections.Sky2Pix_ZenithalPerspective",
    ),
    # astropy.modeling.rotations
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/rotate2d-*"],
        "astropy.modeling.rotations.Rotation2D",
    ),
    Rotate3DConverter(),
    RotationSequenceConverter(),
    # astropy.modeling.tabular
    TabularConverter(),
    # astropy.modeling.bounding_box
    ModelBoundingBoxConverter(),
    CompoundBoundingBoxConverter(),
    SimpleTransformConverter(
        ["tag:stsci.edu:asdf/transform/schechter1d-*"],
        "astropy.modeling.powerlaws.Schechter1D",
    ),
]

# The order here is important; asdf will prefer to use extensions
# that occur earlier in the list.
TRANSFORM_MANIFEST_URIS = [
    # 1.7.0 will be optionally inserted here based on the check below
    "asdf://asdf-format.org/transform/manifests/transform-1.6.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.5.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.4.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.3.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.2.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.1.0",
    "asdf://asdf-format.org/transform/manifests/transform-1.0.0",
]

# Be careful here about including the new transforms.
# Several libraries use a $ref to a specific transform schema
# (old = 1.3.0, new = 1.4.0) to "duck-type" check if something
# is a transform. This is problematic with the old transform-1.3.0
# schema because it also includes specific tag versions for
# anything with a bounding box. For a library that is using the
# 1.3.0 schema if we produce a serialized transform with a new
# bounding box tagged item it will fail validation.
# This is fixed for transform-1.4.0 (which uses a tag wildcard)
# which is included in the 1.7.0 manifest but since the downstream
# libraries need to update references from 1.3.0 to 1.4.0 we need
# to be careful about when we start using the new transform tags.
# Here we check if a few known downstream libraries are installed
# to avoid including the new tags until those libraries can be updated.
# We make the assumption here that the next version of each
# downstream library manifest will fix the issue.
_REQUIRED_DOWNSTREAM_MANIFESTS = {
    "dkist": "asdf://dkist.nso.edu/dkist/extensions/dkist-wcs-1.5.0",
    "asdf_wcs_schemas": "asdf://asdf-format.org/astronomy/gwcs/extensions/gwcs-1.4.0",
    "stdatamodels": "asdf://stsci.edu/jwst_pipeline/extensions/jwst_transforms-1.2.0",
}
_include_new_transforms = True
_resource_manager = asdf.get_config().resource_manager
for package_name, required_manifest_uri in _REQUIRED_DOWNSTREAM_MANIFESTS.items():
    # don't use astropy.utils.minversion as it imports the package
    if find_spec(package_name) is None:
        # not installed
        continue
    if required_manifest_uri not in _resource_manager:
        # don't include new manifest
        _include_new_transforms = False
        break
if _include_new_transforms:
    TRANSFORM_MANIFEST_URIS.insert(0, "asdf://asdf-format.org/transform/manifests/transform-1.7.0")

TRANSFORM_EXTENSIONS = [
    ManifestExtension.from_uri(
        uri,
        # This prevents a warning about a missing extension when opening
        # files written by older versions of the asdf library:
        legacy_class_names=["astropy.io.misc.asdf.extension.AstropyAsdfExtension"],
        converters=TRANSFORM_CONVERTERS,
    )
    for uri in TRANSFORM_MANIFEST_URIS
]


COORDINATES_CONVERTERS = [
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/baseframe-*",
        "astropy.coordinates.baseframe.BaseCoordinateFrame",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/cirs-*",
        "astropy.coordinates.builtin_frames.cirs.CIRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/fk4-*",
        "astropy.coordinates.builtin_frames.fk4.FK4",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/fk4noeterms-*",
        "astropy.coordinates.builtin_frames.fk4.FK4NoETerms",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/fk5-*",
        "astropy.coordinates.builtin_frames.fk5.FK5",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/galactic-*",
        "astropy.coordinates.builtin_frames.galactic.Galactic",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/galactocentric-*",
        "astropy.coordinates.builtin_frames.galactocentric.Galactocentric",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/gcrs-*",
        "astropy.coordinates.builtin_frames.gcrs.GCRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/icrs-1.1.0",
        "astropy.coordinates.builtin_frames.icrs.ICRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/itrs-*",
        "astropy.coordinates.builtin_frames.itrs.ITRS",
    ),
    FrameConverter(
        "tag:astropy.org:astropy/coordinates/frames/precessedgeocentric-*",
        "astropy.coordinates.builtin_frames.gcrs.PrecessedGeocentric",
    ),
    LegacyICRSConverter(),
    AngleConverter(),
    LatitudeConverter(),
    LongitudeConverter(),
    EarthLocationConverter(),
    RepresentationConverter(),
    SkyCoordConverter(),
    SpectralCoordConverter(),
]


ASTROPY_CONVERTERS = [
    UnitsMappingConverter(),
    TimeDeltaConverter(),
    AstropyTableConverter(),
    AstropyFitsConverter(),
    NdarrayMixinConverter(),
    UncertaintyConverter(),
    WCSConverter(),
    SlicedWCSConverter(),
]

_COORDINATES_MANIFEST_URIS = [
    "asdf://asdf-format.org/astronomy/coordinates/manifests/coordinates-1.2.0",
    "asdf://asdf-format.org/astronomy/coordinates/manifests/coordinates-1.1.0",
    "asdf://asdf-format.org/astronomy/coordinates/manifests/coordinates-1.0.0",
]

COORDINATES_EXTENSIONS = [
    ManifestExtension.from_uri(
        manifest_uri,
        converters=COORDINATES_CONVERTERS,
    )
    for manifest_uri in _COORDINATES_MANIFEST_URIS
]


_ASTROPY_EXTENSION_MANIFEST_URIS = [
    "asdf://astropy.org/astropy/manifests/astropy-1.3.0",
    "asdf://astropy.org/astropy/manifests/astropy-1.2.0",
    "asdf://astropy.org/astropy/manifests/astropy-1.1.0",
    "asdf://astropy.org/astropy/manifests/astropy-1.0.0",
]


ASTROPY_EXTENSIONS = [
    ManifestExtension.from_uri(
        manifest_uri,
        # This prevents a warning about a missing extension when opening
        # files written by older versions of astropy:
        legacy_class_names=["astropy.io.misc.asdf.extension.AstropyExtension"],
        converters=ASTROPY_CONVERTERS,
    )
    for manifest_uri in _ASTROPY_EXTENSION_MANIFEST_URIS
]


# These tags are part of the ASDF Standard,
# but we want to override serialization here so that users can
# work with nice astropy objects for those entities.
_FITS_CONVERTERS = [
    AsdfFitsConverter(),
]

_TIME_CONVERTERS = [
    TimeConverter(),
]

_TABLE_CONVERTERS = [
    ColumnConverter(),
    AsdfTableConverter(),
]

_UNIT_CONVERTERS = [
    UnitConverter(),
    EquivalencyConverter(),
    MagUnitConverter(),
    QuantityConverter(),
]

CORE_CONVERTERS = _FITS_CONVERTERS + _TIME_CONVERTERS + _TABLE_CONVERTERS + _UNIT_CONVERTERS

UNIT_EXTENSIONS = [
    ManifestExtension.from_uri(
        "asdf://astropy.org/astropy/manifests/units-1.0.0",
        converters=_UNIT_CONVERTERS,
    ),
]

# up to asdf 1.5.0 many tags supported by asdf-astropy
# were defined in core manifests
CORE_MANIFEST_URIS = [
    "asdf://asdf-format.org/astronomy/manifests/astronomy-1.0.0",
    "asdf://asdf-format.org/core/manifests/core-1.5.0",
    "asdf://asdf-format.org/core/manifests/core-1.4.0",
    "asdf://asdf-format.org/core/manifests/core-1.3.0",
    "asdf://asdf-format.org/core/manifests/core-1.2.0",
    "asdf://asdf-format.org/core/manifests/core-1.1.0",
    "asdf://asdf-format.org/core/manifests/core-1.0.0",
]

if importlib.metadata.version("asdf-standard") > "1.2.0":
    CORE_MANIFEST_URIS.insert(0, "asdf://asdf-format.org/astronomy/manifests/astronomy-1.1.0")

CORE_EXTENSIONS = [ManifestExtension.from_uri(u, converters=CORE_CONVERTERS) for u in CORE_MANIFEST_URIS]


# asdf-astropy 0.4.0 combined core and unit manifests
# but registered them with the core uri
# asdf-astropy 0.5.0 combined core and unit manifests
# but registered them with the core uri with asdf-format.org
# swapped with astropy.org
# To avoid warnings for opening old files we register empty
# extensions with the astropy.org uris here
class _EmptyExtension(Extension):
    def __init__(self, uri):
        self._uri = uri

    @property
    def extension_uri(self):
        return self._uri


EMPTY_EXTENSIONS = [
    _EmptyExtension(uri)
    for uri in [
        "asdf://astropy.org/core/extensions/core-1.5.0",
        "asdf://astropy.org/core/extensions/core-1.4.0",
        "asdf://astropy.org/core/extensions/core-1.3.0",
        "asdf://astropy.org/core/extensions/core-1.2.0",
        "asdf://astropy.org/core/extensions/core-1.1.0",
        "asdf://astropy.org/core/extensions/core-1.0.0",
    ]
]



import asdf
import numpy as np
import pytest
from asdf.testing import helpers
from astropy import units
from astropy.units import Quantity
from astropy.utils.introspection import minversion
from numpy.testing import assert_array_equal

from asdf_astropy.tests.versions import ASTROPY_GE_7_1


def asdf_open_memory_mapping_kwarg(memmap: bool) -> dict:
    if minversion("asdf", "3.1.0"):
        return {"memmap": memmap}
    return {"copy_arrays": not memmap}


def create_quantities():
    return [
        # Scalar:
        Quantity(2.71828, units.kpc),
        # Non-float scalar:
        Quantity(7, units.K, dtype=np.int32),
        # Single element array:
        Quantity([3.14159], units.kg),
        # Multiple element array:
        Quantity([x * 2.3081 for x in range(10)], units.ampere),
        # Multiple dimension array:
        Quantity(np.arange(100, dtype=np.float64).reshape(5, 20), units.km),
        # Non-float array:
        Quantity(np.zeros((5, 5), dtype=np.uint16), units.cm, dtype=np.uint16),
    ]


@pytest.mark.parametrize("quantity", create_quantities())
def test_serialization(quantity, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["quantity"] = quantity
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert (af["quantity"].value == quantity.value).all()
        assert af["quantity"].dtype == quantity.dtype
        assert (af["quantity"] == quantity).all()


def test_read_untagged_unit():
    value = 2.71828
    yaml = f"""
quantity: !unit/quantity-1.1.0
  value: {value}
  unit: kpc
    """
    buff = helpers.yaml_to_asdf(yaml, version="1.5.0")
    with asdf.open(buff) as af:
        assert af["quantity"].value == value
        assert af["quantity"].unit.is_equivalent(units.kpc)


def test_read_tagged_unit():
    value = 2.71828
    yaml = f"""
quantity: !unit/quantity-1.1.0
  value: {value}
  unit: !unit/unit-1.0.0 kpc
    """
    buff = helpers.yaml_to_asdf(yaml, version="1.5.0")
    with asdf.open(buff) as af:
        assert af["quantity"].value == value
        assert af["quantity"].unit.is_equivalent(units.kpc)


def test_read_array_value():
    yaml = """
quantity: !unit/quantity-1.1.0
  value: !core/ndarray-1.0.0 [1.0, 2.0, 3.0, 4.0]
  unit: km
    """
    buff = helpers.yaml_to_asdf(yaml, version="1.5.0")
    with asdf.open(buff) as af:
        assert_array_equal(af["quantity"].value, np.array([1.0, 2.0, 3.0, 4.0]))
        assert af["quantity"].unit.is_equivalent(units.km)


def test_memmap(tmp_path):
    """
    Test that memmap (memmap=True) works with quantities.

    Unfortunately, this is not a simple `isinstance(obj, np.memmap)`
    Instead it requires a more complicated check.
    """
    file_path = tmp_path / "test.asdf"
    quantity = Quantity(np.arange(100, dtype=np.float64).reshape(5, 20), units.km)

    new_value = 42.14 * units.cm
    new_quantity = quantity.copy()
    new_quantity[-1, -1] = new_value

    # Write initial ASDF file
    with asdf.AsdfFile() as af:
        af.tree = {"quantity": quantity}
        af.write_to(file_path)

    # Update a value in the ASDF file
    with asdf.open(file_path, mode="rw", **asdf_open_memory_mapping_kwarg(memmap=True)) as af:
        assert (af.tree["quantity"] == quantity).all()
        assert af.tree["quantity"][-1, -1] != new_value

        af.tree["quantity"][-1, -1] = new_value

        assert af.tree["quantity"][-1, -1] == new_value
        assert (af.tree["quantity"] != quantity).any()
        assert (af.tree["quantity"] == new_quantity).all()

    with asdf.open(file_path, mode="rw", **asdf_open_memory_mapping_kwarg(memmap=True)) as af:
        assert af.tree["quantity"][-1, -1] == new_value
        assert (af.tree["quantity"] != quantity).any()
        assert (af.tree["quantity"] == new_quantity).all()


def test_no_memmap(tmp_path):
    """
    Test that turning off memmap (memmap=False) works as expected for quantities
    """
    file_path = tmp_path / "test.asdf"
    quantity = Quantity(np.arange(100, dtype=np.float64).reshape(5, 20), units.km)

    new_value = 42.14 * units.cm
    new_quantity = quantity.copy()
    new_quantity[-1, -1] = new_value

    # Write initial ASDF file
    with asdf.AsdfFile() as af:
        af.tree = {"quantity": quantity}
        af.write_to(file_path)

    # Update a value in the ASDF file
    with asdf.open(file_path, mode="rw", **asdf_open_memory_mapping_kwarg(memmap=False)) as af:
        assert (af.tree["quantity"] == quantity).all()
        assert af.tree["quantity"][-1, -1] != new_value

        af.tree["quantity"][-1, -1] = new_value

        assert af.tree["quantity"][-1, -1] == new_value
        assert (af.tree["quantity"] != quantity).any()
        assert (af.tree["quantity"] == new_quantity).all()

    with asdf.open(file_path, mode="rw", **asdf_open_memory_mapping_kwarg(memmap=False)) as af:
        assert af.tree["quantity"][-1, -1] != new_value
        assert (af.tree["quantity"] != new_quantity).any()
        assert (af.tree["quantity"] == quantity).all()


@pytest.mark.skipif(ASTROPY_GE_7_1, reason="MaskedQuantity support was added in astropy 7.1")
def test_masked_quantity_raises():
    yaml = """
quantity: !unit/quantity-1.1.0
  unit: !unit/unit-1.0.0 Ymol
  value: !core/ndarray-1.0.0
    data: [1.0, 2.0, null]
    mask: !core/ndarray-1.0.0
      data: [false, false, true]
      datatype: bool8
      shape: [3]
    datatype: float64
    shape: [3]
"""
    buff = helpers.yaml_to_asdf(yaml, version="1.5.0")
    with pytest.raises(NotImplementedError, match="MaskedQuantity support requires astropy 7.1 or later"):
        asdf.open(buff)



import asdf
import numpy as np
import pytest
from astropy import units as u
from astropy.nddata import StdDevUncertainty, UnknownUncertainty, VarianceUncertainty


def create_uncertainty():
    uncert = np.arange(100).reshape(10, 10)
    uncertainty_stddev_1 = StdDevUncertainty(uncert, unit="m")
    uncertainty_stddev_2 = StdDevUncertainty([2], unit="m")
    uncertainty_unknown_1 = UnknownUncertainty(uncert, unit="m")
    uncertainty_unknown_2 = UnknownUncertainty([0.4], unit=u.adu)
    uncertainty_variance_1 = VarianceUncertainty(uncert, unit="m")
    uncertainty_variance_2 = VarianceUncertainty([0.4], unit=u.adu)
    return [
        uncertainty_stddev_1,
        uncertainty_stddev_2,
        uncertainty_unknown_1,
        uncertainty_unknown_2,
        uncertainty_variance_1,
        uncertainty_variance_2,
    ]


@pytest.mark.parametrize("uncertainty", create_uncertainty())
def test_uncertainty_serialization(uncertainty, tmp_path):
    file_path = tmp_path / "test_uncertainty.asdf"
    with asdf.AsdfFile() as af:
        af["uncertainty"] = uncertainty
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        loaded_uncert = af["uncertainty"]
        assert type(loaded_uncert) is type(uncertainty)
        assert (loaded_uncert._array == uncertainty._array).all()
        assert loaded_uncert.unit == uncertainty.unit



import asdf
import numpy as np
import pytest
from astropy.wcs import WCS
from astropy.wcs.wcsapi.wrappers.sliced_wcs import SlicedLowLevelWCS

from asdf_astropy.testing.helpers import assert_wcs_equal


def create_wcs():
    wcs = WCS(naxis=4)
    wcs.wcs.ctype = "RA---CAR", "DEC--CAR", "FREQ", "TIME"
    wcs.wcs.cunit = "deg", "deg", "Hz", "s"
    wcs.wcs.cdelt = -2.0, 2.0, 3.0e9, 1
    wcs.wcs.crval = 4.0, 0.0, 4.0e9, 3
    wcs.wcs.crpix = 6.0, 7.0, 11.0, 11.0
    wcs.wcs.cname = "Right Ascension", "Declination", "Frequency", "Time"

    wcs0 = SlicedLowLevelWCS(wcs, 1)
    wcs1 = SlicedLowLevelWCS(wcs, [slice(None), slice(None), slice(None), 10])
    wcs3 = SlicedLowLevelWCS(SlicedLowLevelWCS(wcs, slice(None)), [slice(3), slice(None), slice(None), 10])
    wcs_ellipsis = SlicedLowLevelWCS(wcs, [Ellipsis, slice(5, 10)])
    wcs2 = SlicedLowLevelWCS(wcs, np.s_[:, 2, 3, :])
    return [wcs0, wcs1, wcs2, wcs_ellipsis, wcs3]


@pytest.mark.parametrize("sl_wcs", create_wcs())
def test_sliced_wcs_serialization(sl_wcs, tmp_path):
    file_path = tmp_path / "test_slicedwcs.asdf"
    with asdf.AsdfFile() as af:
        af["sl_wcs"] = sl_wcs
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        loaded_sl_wcs = af["sl_wcs"]
        assert_wcs_equal(sl_wcs._wcs, loaded_sl_wcs._wcs)
        assert sl_wcs._slices_array == loaded_sl_wcs._slices_array



from asdf.extension import Converter
from asdf.tags.core.ndarray import NDArrayType


class QuantityConverter(Converter):
    tags = ("tag:stsci.edu:asdf/unit/quantity-*",)
    types = (
        "astropy.units.quantity.Quantity",
        # The Distance class has no tag of its own, so we
        # just serialize it as a quantity.
        "astropy.coordinates.distances.Distance",
        "astropy.utils.masked.core.MaskedQuantity",
    )

    def to_yaml_tree(self, obj, tag, ctx):
        import numpy as np
        from astropy.utils.masked import Masked

        node = {
            "value": np.ma.asarray(obj.value) if isinstance(obj, Masked) else obj.value,
            "unit": obj.unit,
        }

        if obj.isscalar:
            node["datatype"] = obj.dtype.name

        return node

    def from_yaml_tree(self, node, tag, ctx):
        # numpy 2.0 changed behavior for copy where an error is produced
        # if False and a copy is required (previously there was no error)
        # astropy 6.1 changed Quantity in a similar way
        import numpy as np
        from astropy.units import Quantity

        copy = None if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1" else False

        value = node["value"]
        dtype = node.get("datatype", None)
        if isinstance(value, NDArrayType):
            # TODO: Why doesn't NDArrayType work?  This needs some research
            # and documentation.
            value = value._make_array()
            dtype = value.dtype

        if isinstance(value, np.ma.MaskedArray):
            from astropy.utils import minversion
            from astropy.utils.masked import Masked

            if not minversion("astropy", "7.1.dev"):  # ASTROPY_GE_7_1
                msg = "MaskedQuantity support requires astropy 7.1 or later"
                raise NotImplementedError(msg)
            class_ = Masked(Quantity)
        else:
            class_ = Quantity

        return class_(value, unit=node["unit"], copy=copy, dtype=dtype)



from asdf.extension import Converter


class SlicedWCSConverter(Converter):
    tags = ("tag:astropy.org:astropy/wcs/slicedwcs-*",)
    types = ("astropy.wcs.wcsapi.wrappers.sliced_wcs.SlicedLowLevelWCS",)

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.wcs.wcsapi.wrappers.sliced_wcs import SlicedLowLevelWCS

        wcs = node["wcs"]
        slice_array = [
            s if isinstance(s, int) else slice(s["start"], s["stop"], s["step"]) for s in node["slices_array"]
        ]
        return SlicedLowLevelWCS(wcs, slice_array)

    def to_yaml_tree(self, sl, tag, ctx):
        slices_array = []

        for s in sl._slices_array:
            if isinstance(s, slice):
                slices_array.append(
                    {
                        "start": s.start,
                        "stop": s.stop,
                        "step": s.step,
                    },
                )
            else:
                slices_array.append(s)
        return {
            "wcs": sl._wcs,
            "slices_array": slices_array,
        }



import asdf
import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import Angle, EarthLocation, Latitude, Longitude, SkyCoord
from astropy.utils.masked import Masked

from asdf_astropy.tests.versions import ASTROPY_GE_7_1

if ASTROPY_GE_7_1:
    from astropy.utils.masked import get_data_and_mask
else:
    pytest.skip(reason="MaskedQuantity support was added in astropy 7.1", allow_module_level=True)


@pytest.mark.parametrize("angle_class", [Angle, Latitude, Longitude])
def test_masked_angle(tmp_path, angle_class):
    angle = [1, 2, 3]
    mask = [False, False, True]
    masked_class = Masked(angle_class)
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["angle"] = Masked(angle_class(angle * u.deg), mask)
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert isinstance(af["angle"], masked_class)
        out_data, out_mask = get_data_and_mask(af["angle"].to_value(u.deg))
        np.testing.assert_array_equal(out_data, angle)
        np.testing.assert_array_equal(out_mask, mask)


def test_masked_skycoord(tmp_path):
    ra_deg = [0.0, 1.0, 2.0]
    dec_deg = [2.0, 1.0, 0.0]
    mask = [False, False, True]
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["coord"] = SkyCoord(Masked(ra_deg, mask), Masked(dec_deg, mask), unit=u.deg)
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert af["coord"].ra.unit == u.deg
        assert af["coord"].dec.unit == u.deg

        out_data, out_mask = get_data_and_mask(af["coord"].ra.deg)
        np.testing.assert_array_equal(out_data, ra_deg)
        np.testing.assert_array_equal(out_mask, mask)
        out_data, out_mask = get_data_and_mask(af["coord"].dec.deg)
        np.testing.assert_array_equal(out_data, dec_deg)
        np.testing.assert_array_equal(out_mask, mask)


def test_masked_earth_location(tmp_path):
    x = [0.0, 1.0, 2.0]
    y = [2.0, 1.0, 0.0]
    z = [1.0, 2.0, 0.0]
    mask = [False, False, True]
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["loc"] = EarthLocation.from_geocentric(Masked(x * u.m, mask), Masked(y * u.m, mask), Masked(z * u.m, mask))
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert af["loc"].x.unit == u.m
        assert af["loc"].y.unit == u.m
        assert af["loc"].z.unit == u.m

        out_data, out_mask = get_data_and_mask(af["loc"].x.to_value(u.m))
        np.testing.assert_array_equal(out_data, x)
        np.testing.assert_array_equal(out_mask, mask)
        out_data, out_mask = get_data_and_mask(af["loc"].y.to_value(u.m))
        np.testing.assert_array_equal(out_data, y)
        np.testing.assert_array_equal(out_mask, mask)
        out_data, out_mask = get_data_and_mask(af["loc"].z.to_value(u.m))
        np.testing.assert_array_equal(out_data, z)
        np.testing.assert_array_equal(out_mask, mask)
